"""
Basic types we will more widely need in DaCHS;  we also import everything
we need from typing, so for our type annotations we only need to say
from utils.dachstypes import ...

This also lets us smuggle in fallbacks for python 3.7 (and we won't go
further back than that).
"""

#c Copyright 2008-2025, the GAVO project <gavo@ari.uni-heidelberg.de>
#c
#c This program is free software, covered by the GNU GPL.  See the
#c COPYING file in the source distribution.

from typing import ( #noflake: gratuitous imports for re-exporting
	AbstractSet, Any,
	BinaryIO,
	Callable, ClassVar, ContextManager,
	Dict,
	Final,
	FrozenSet,
	Generator,
	Hashable,
	IO, Iterable, Iterator,
	Generic,
	Hashable,
	List,
	Mapping,
	NewType, NoReturn,
	Optional,
	Protocol,
	Sequence, Set,
	TextIO, Tuple, TYPE_CHECKING, Type, TypeVar,
	Union,
	cast)

# things introduced after python 3.7; if their import fails, we replace
# them with None, so type checking will fail with older pythons, but
# that seems fine to me.
try:
	from typing import ( #noflake: gratuitous imports for re-exporting
		Self
	)
except ImportError:
	Self = None  # type: ignore

from types import ModuleType  #noflake: gratuitous imports for re-exporting
from re import Match  #noflake: gratuitous imports for re-exporting

import numpy

try:
	from numpy.typing import NDArray
except ImportError:
	NDArray = NewType('NDArray', object)  # type: ignore

Numeric = Union[int, float, numpy.float64]

# a base.events event handler
BaseEventHandler = Callable[..., None]

import pathlib
Filename = Union[str, pathlib.Path]
# positions when doing XML parsing
ParsePosition =  NewType('ParsePosition', object)
StrOrBytes = Union[str, bytes]
StrToStrMap = Dict[str, str]

# see utils.plainxml
PlainXMLEvent = Tuple[
	str,
	Optional[str],
	Union[Dict[str, str], None, str]]

if TYPE_CHECKING:
	from gavo.base import Structure
else:
	Structure = None # type: ignore

# base.common wants a structure parse context, i.e., something like
# ParseContext or RDParseContext.
# TODO: this could do with better type annotations and more functions.
class StructParseContext(Protocol):
	restricted: bool = True
	replayTarget: "Structure"

	def addExitFunc(self, callable: Callable[[], None]) -> None:
		pass
	
	@property
	def pos(self) -> Any:
		pass
	
	def setPositionOn(self, struct: Structure):
		pass
	
	def getById(self,
			id: str,
			forceType=Optional[type]) -> Structure:
		pass

	def resolveId(self,
			id: str,
			instance: Optional[str]=None,
			forceType=Optional[type]) -> Structure:
		pass

	def replaying(self) -> Any:
		pass


# the type of value arguments in struct parser events.
StructParserValue = Optional[Union[Dict[str, str], str]]

# One of these days we ought to be a bit more concrete about which
# source tokens we accept; but it's really up to the grammars...
SourceToken = Any
